from django.shortcuts import get_object_or_404
from rest_framework.generics import \
    ListAPIView, ListCreateAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.views import APIView
from rest_framework.parsers import MultiPartParser, FileUploadParser
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework import status
from rest_framework.exceptions import ValidationError, NotAuthenticated, AuthenticationFailed
from arkindex.documents.models import Corpus
from arkindex.dataimport.models import \
    DataImport, DataFile, DataImportState, DataImportMode, Repository, RepositorySource, Revision
from arkindex.dataimport.serializers import \
    DataImportLightSerializer, DataImportSerializer, DataFileSerializer
import hashlib
import magic


class DataImportsList(ListCreateAPIView):
    """
    List data imports
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportLightSerializer

    def get_queryset(self):
        return DataImport.objects.filter(creator=self.request.user)

    def perform_create(self, serializer):
        # Temporary limitation
        if serializer.validated_data['mode'] not in (DataImportMode.Images, ):
            raise ValidationError('Unsupported mode for now, sorry.')

        return super().perform_create(serializer)


class DataImportDetails(RetrieveUpdateDestroyAPIView):
    """
    Retrieve and edit a data import
    """

    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportSerializer

    def get_queryset(self):
        return DataImport.objects.filter(creator=self.request.user)

    def perform_update(self, serializer):
        dataimport = serializer.instance
        if dataimport.state not in (DataImportState.Created, DataImportState.Configured):
            raise ValidationError("Cannot update a workflow after it has been started")

        # Allow updating the state from Configured to Running only
        if dataimport.state != serializer.validated_data['state']:
            if dataimport.state != DataImportState.Configured:
                raise ValidationError({'state': ['Cannot change state on non-configured or started workflows']})
            if serializer.validated_data['state'] != DataImportState.Running:
                raise ValidationError(
                    {'state': ['The state of a configured workflow can only be set to configured or running']})
            dataimport.start()
            return

        super().perform_update(serializer)

        # Automatically change from Created to Configured if a payload has been set
        if dataimport.state == DataImportState.Created and dataimport.payload is not None:
            dataimport.state = DataImportState.Configured
            dataimport.save()

    def perform_destroy(self, instance):
        if instance.state == DataImportState.Running:
            raise ValidationError("Cannot delete a workflow while it is running")
        super().perform_destroy(instance)


class DataFileList(ListAPIView):
    """
    List uploaded files
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        return DataFile.objects.filter(corpus_id=self.kwargs['pk'])


class DataFileRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataFileSerializer
    # TODO: Restrict to user's corpuses
    queryset = DataFile.objects.all()


class DataFileUpload(APIView):
    """
    Upload a new file to a corpus
    """
    permission_classes = (IsAuthenticated, )
    parser_classes = (MultiPartParser, FileUploadParser)

    def post(self, request, pk=None, format=None):
        corpus_qs = Corpus.objects.filter(id=pk)
        if not corpus_qs.exists():
            raise ValidationError({'corpus': ['Corpus not found']})
        corpus = corpus_qs.get()

        file_obj = request.FILES['file']

        md5hash = hashlib.md5()
        for chunk in file_obj.chunks():
            md5hash.update(chunk)
        file_hash = md5hash.hexdigest()

        # Reopen file to reread from beginning
        file_obj.open()
        file_type = magic.from_buffer(file_obj.read(1024), mime=True)

        if DataFile.objects.filter(hash=file_hash, corpus=corpus).exists():
            raise ValidationError('File already exists')

        df = DataFile.objects.create(
            hash=file_hash,
            corpus=corpus,
            name=file_obj.name,
            size=file_obj.size,
            content_type=file_type,
        )

        file_obj.open()
        with open(df.staging_path, 'wb') as destfile:
            for chunk in file_obj.chunks():
                destfile.write(chunk)

        return Response(
            data=DataFileSerializer(df).data,
            status=status.HTTP_201_CREATED)


class GitRepositoryImportHook(APIView):
    """
    Handle Git push events
    """

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)

        if repo.source == RepositorySource.GitLab:
            if 'HTTP_X_GITLAB_EVENT' not in request.META:
                raise ValidationError("Missing GitLab event type")
            if request.META['HTTP_X_GITLAB_EVENT'] != 'Push Hook':
                raise ValidationError("Unsupported GitLab event type")

            if 'HTTP_X_GITLAB_TOKEN' not in request.META:
                raise NotAuthenticated("Missing GitLab secret token")
            if request.META['HTTP_X_GITLAB_TOKEN'] != repo.hook_token:
                raise AuthenticationFailed("Invalid GitLab secret token")

            assert isinstance(request.data, dict)
            assert request.data['object_kind'] == 'push'

            if request.data['ref'] not in repo.watched_branches:
                return Response(status=status.HTTP_204_NO_CONTENT)

            # Already took care of this event
            if Revision.objects.filter(
                    repo=repo,
                    ref=request.data['ref'],
                    hash=request.data['checkout_sha']).exists():
                return Response(status=status.HTTP_204_NO_CONTENT)

            rev = Revision.objects.create(
                repo=repo,
                hash=request.data['checkout_sha'],
                ref=request.data['ref'],
                message=request.data['commits'][-1]['message'],
                author=request.data['commits'][-1]['author']['name'],
            )
        else:
            raise NotImplementedError

        rev.start_import()
        return Response(status=status.HTTP_204_NO_CONTENT)
