from celery import shared_task
from celery.utils.log import get_task_logger
from celery.signals import task_postrun
from celery.states import EXCEPTION_STATES
from django.conf import settings
from django.db import transaction
from arkindex.project.celery import ReportingTask
from arkindex.documents.models import Element, ElementType
from arkindex.documents.importer import import_page
from arkindex.documents.tei import TeiParser
from arkindex.images.models import ImageServer, ImageStatus
from arkindex.dataimport.models import DataImport, DataImportState, DataImportMode
from PIL import Image
import os
import glob
import logging
import shutil
import git
import urllib.parse

root_logger = logging.getLogger(__name__)
logger = get_task_logger(__name__)


class TaskLoggingHandler(logging.Handler):

    def __init__(self, task, level=logging.WARNING):
        assert isinstance(task, ReportingTask)
        super().__init__(level=level)
        self.task = task

    def emit(self, record):
        self.task.report_message(record.getMessage(), level=record.levelno)


@shared_task(bind=True, base=ReportingTask)
def check_images(self, dataimport):
    assert isinstance(dataimport, DataImport)

    datafiles = dataimport.files.all()
    valid_files = []
    filecount = len(datafiles)

    for i, datafile in enumerate(datafiles):
        self.report_progress(i / filecount, "Checking image {} of {}".format(i + 1, filecount))
        fmt = None

        try:
            fmt = Image.open(datafile.staging_path).format
            assert fmt in ('JPEG', 'JPEG2000')
        except IOError as e:
            self.report_message("File {} is not a valid image".format(datafile.name), level=logging.WARNING)
            continue
        except AssertionError:
            self.report_message("File {} is not a valid JPEG image".format(datafile.name), level=logging.WARNING)
            continue

        valid_files.append(datafile)

    assert len(valid_files) > 0, "No valid JPEG images in selected files"
    return valid_files


@shared_task(bind=True, base=ReportingTask)
def import_images(self, valid_files, dataimport, server_id=settings.LOCAL_IMAGESERVER_ID):
    assert isinstance(dataimport, DataImport)
    self.report_progress(0, 'Pre import checks...')
    try:
        server = ImageServer.objects.get(id=server_id)
    except ImageServer.DoesNotExist:
        raise ValueError('Image server {} does not exist'.format(server_id))

    if os.path.exists(dataimport.iiif_path):
        raise ValueError('Folder with name {} already exists'.format(dataimport.folder_name))
    os.makedirs(dataimport.iiif_path)

    # Get volume
    volume_name = dataimport.payload['volume_name']
    vol, _ = Element.objects.get_or_create(
        type=ElementType.Volume,
        name=volume_name,
        corpus=dataimport.corpus,
    )

    datafiles = dataimport.files.all()
    for i, datafile in enumerate(datafiles, 1):
        self.report_progress(i / len(datafiles), 'Importing image {} of {}'.format(i, len(datafiles)))

        pillow_img = Image.open(datafile.staging_path)
        width, height = pillow_img.size

        ext = '.jp2' if pillow_img.format == 'JPEG2000' else '.jpg'
        newfilename = str(datafile.id) + ext
        shutil.copyfile(datafile.staging_path, os.path.join(dataimport.iiif_path, newfilename))

        img, _ = server.images.get_or_create(
            path=urllib.parse.urljoin(dataimport.folder_name + '/', newfilename),
            defaults={
                'width': width,
                'height': height,
                'status': ImageStatus.Checked,
            }
        )

        import_page(vol, img, volume_name, str(i), i)

    self.report_message("Imported files into {}".format(vol))
    return {'volume': str(vol.id)}


@shared_task(bind=True, base=ReportingTask)
def clone_repo(self, dataimport):
    assert isinstance(dataimport, DataImport)
    assert dataimport.mode == DataImportMode.Repository
    assert dataimport.revision is not None

    self.report_progress(0, "Cloning repository...")
    repo_dir = dataimport.revision.repo.clone_dir
    if os.path.exists(repo_dir):
        shutil.rmtree(repo_dir)
    repo = git.Repo.clone_from(dataimport.revision.repo.auth_url, repo_dir, no_checkout=True)

    commit_hash = dataimport.revision.hash
    self.report_progress(0.5, "Checking out commit {}...".format(commit_hash))
    repo.head.reference = repo.create_head('commit_{}'.format(commit_hash), commit_hash)
    repo.head.reset(index=True, working_tree=True)


@shared_task(bind=True, base=ReportingTask)
def import_repo(self, dataimport):
    handler = TaskLoggingHandler(self)
    root_logger.addHandler(handler)

    self.report_progress(0, "Finding XML files...")
    xml_files = glob.glob(os.path.join(dataimport.revision.repo.clone_dir, '**/*.xml'), recursive=True)

    for i, xml_file in enumerate(xml_files, 1):
        filename = os.path.basename(xml_file)
        self.report_progress(i / len(xml_files), 'Importing file {} of {}: {}'.format(i, len(xml_files), filename))

        try:
            parser = TeiParser(xml_file)
            parser.check()
            matches = parser.match_database(dataimport.corpus)

            for db_elt, tei_elt in matches:
                with transaction.atomic():
                    tei_elt.save(db_elt, dataimport.revision)

        except Exception as e:
            self.report_message(
                "Importing of {} failed: {}".format(filename, str(e)), level=logging.WARNING)

    root_logger.removeHandler(handler)


@shared_task(bind=True, base=ReportingTask)
def cleanup_repo(self, dataimport):
    shutil.rmtree(dataimport.revision.repo.clone_dir)


@task_postrun.connect
def dataimport_postrun(task_id, task, state, args=(), **kwargs):
    '''
    Update DataImport state after task exit
    '''

    # Only for dataimport tasks
    if not task.name.startswith('arkindex.dataimport.tasks'):
        return

    # Look for dataimport in args
    imps = [a for a in args if isinstance(a, DataImport)]
    assert len(imps) == 1, 'No args on dataimport task.'
    dataimport = imps[0]
    dataimport.refresh_from_db()  # avoid inconsistency
    assert isinstance(dataimport, DataImport), \
        'DataImport not found as first arg of task {}'.format(task)

    # Manage error state
    if any(t.failed() for t in dataimport.tasks) or state in EXCEPTION_STATES:
        dataimport.state = DataImportState.Error

        # Report exception message
        if 'retval' in kwargs:
            task.report_message(str(kwargs['retval']), logging.ERROR)

    # Manage completion when all tasks are completed
    elif sum(t.successful() for t in dataimport.tasks) == dataimport.task_count:
        dataimport.state = DataImportState.Done

    dataimport.save()
