from rest_framework import serializers
from arkindex.documents.models import \
    Element, ElementType, Page, PageType, PageDirection, Act, Corpus, MetaData, MetaType
from arkindex.images.serializers import ZoneSerializer, ImageSerializer
from arkindex.dataimport.serializers import RevisionSerializer
from arkindex.project.serializer_fields import EnumField, ViewerURLField


class MetaDataSerializer(serializers.ModelSerializer):
    """
    Serialises some Metadata for any Element
    """
    type = EnumField(MetaType)
    revision = RevisionSerializer()

    class Meta:
        model = MetaData
        fields = (
            'id',
            'type',
            'name',
            'value',
            'revision',
        )


class ElementLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Element
    """
    type = EnumField(ElementType)

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
        )


class PageLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Page
    """
    page_type = EnumField(PageType)
    direction = EnumField(PageDirection)
    image = ImageSerializer(source='zone.image')

    class Meta:
        model = Page
        fields = (
            'id',
            'page_type',
            'nb',
            'direction',
            'display_name',
            'image',
        )


class ElementSerializer(serializers.ModelSerializer):
    """
    Fully Serialises a document
    """
    type = EnumField(ElementType)

    # TODO: detect correct manifest per element type
    viewer_url = ViewerURLField('api:volume-manifest')

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
            'viewer_url',
        )


class ActSerializer(serializers.ModelSerializer):
    """
    Serialize an act with its parents and children
    and metadatas
    """

    parents = serializers.ListField(
        child=serializers.ListField(
            child=ElementLightSerializer()
        ),
        source='parent_elements',
        read_only=True,
    )
    children = ElementLightSerializer(source='child_elements', read_only=True, many=True)
    metadatas = MetaDataSerializer(read_only=True, many=True)

    class Meta:
        model = Act
        fields = (
            'id',
            'name',
            'number',
            'parents',
            'children',
            'metadatas',
        )
        read_only_fields = ('id', )


class SurfaceSerializer(serializers.ModelSerializer):
    """
    Serialize a surface with its page, image and zone
    """

    zone = ZoneSerializer()
    page = PageLightSerializer()

    class Meta:
        model = Element
        fields = (
            'id',
            'name',
            'page',
            'zone',
        )


class CorpusSerializer(serializers.ModelSerializer):
    """
    Serialize a corpus
    """

    class Meta:
        model = Corpus
        fields = ('id', 'name')
